@@ -99,6 +99,8 @@ struct SDParams {
     bool clip_on_cpu              = false;
     bool vae_on_cpu               = false;
     bool diffusion_flash_attn     = false;
+    bool diffusion_conv_direct    = false;
+    bool vae_conv_direct          = false;
     bool canny_preprocess         = false;
     bool color                    = false;
     int upscale_repeats           = 1;
@@ -134,6 +136,27 @@ static bool sd_is_quiet = false;
 static std::string sdmodelfilename = "";
 static bool photomaker_enabled = false;
 
+static void set_sd_vae_tiling(sd_ctx_t* ctx, bool tiling)
+{
+    ctx->sd->vae_tiling = tiling;
+}
+
+static int get_loaded_sd_version(sd_ctx_t* ctx)
+{
+    return ctx->sd->version;
+}
+
+static bool loaded_model_is_chroma(sd_ctx_t* ctx)
+{
+    if (ctx != nullptr && ctx->sd != nullptr) {
+        auto maybe_flux = std::dynamic_pointer_cast<FluxModel>(ctx->sd->diffusion_model);
+        if (maybe_flux != nullptr) {
+            return maybe_flux->flux.flux_params.is_chroma;
+        }
+    }
+    return false;
+}
+
 bool sdtype_load_model(const sd_load_model_inputs inputs) {
     sd_is_quiet = inputs.quiet;
     set_sd_quiet(sd_is_quiet);
@@ -160,6 +183,10 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
     {
         taesdpath = executable_path + "taesd.embd";
         printf("With TAE SD VAE: %s\n",taesdpath.c_str());
+        if (cfg_tiled_vae_threshold < 8192) {
+            printf("  disabling VAE tiling for TAESD\n");
+            cfg_tiled_vae_threshold = 8192;
+        }
     }
     else if(vaefilename!="")
     {
@@ -186,7 +213,15 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
     {
         printf("Flash Attention is enabled\n");
     }
-    if(inputs.quant)
+    if(inputs.diffusion_conv_direct)
+    {
+        printf("Conv2D Direct for diffusion model is enabled\n");
+    }
+    if(inputs.vae_conv_direct)
+    {
+        printf("Conv2D Direct for VAE model is enabled\n");
+    }
+    if(inputs.quant > 0)
     {
         printf("Note: Loading a pre-quantized model is always faster than using compress weights!\n");
     }
@@ -218,9 +253,15 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
 
     sd_params = new SDParams();
     sd_params->model_path = inputs.model_filename;
-    sd_params->wtype = (inputs.quant==0?SD_TYPE_COUNT:SD_TYPE_Q4_0);
+    sd_params->wtype = SD_TYPE_COUNT;
+    if (inputs.quant > 0) {
+        sd_params->wtype = (inputs.quant==1?SD_TYPE_Q8_0:SD_TYPE_Q4_0);
+        printf("Diffusion Model quantized to %s\n", sd_type_name(sd_params->wtype));
+    }
     sd_params->n_threads = inputs.threads; //if -1 use physical cores
     sd_params->diffusion_flash_attn = inputs.flash_attention;
+    sd_params->diffusion_conv_direct = inputs.diffusion_conv_direct;
+    sd_params->vae_conv_direct = inputs.vae_conv_direct;
     sd_params->input_path = ""; //unused
     sd_params->batch_count = 1;
     sd_params->vae_path = vaefilename;
@@ -233,9 +274,13 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
     bool endswithgguf = (sd_params->model_path.rfind(".gguf") == sd_params->model_path.size() - 5);
     if(sd_params->t5xxl_path!="" && endswithgguf)
     {
-        printf("\nSwap to Diffusion Model Path:%s",sd_params->model_path.c_str());
-        sd_params->diffusion_model_path = sd_params->model_path;
-        sd_params->model_path = "";
+        //extra check - make sure there is no diffusion model prefix already inside!
+        if(!gguf_tensor_exists(sd_params->model_path,"model.diffusion_model.",false))
+        {
+            printf("\nSwap to Diffusion Model Path:%s",sd_params->model_path.c_str());
+            sd_params->diffusion_model_path = sd_params->model_path;
+            sd_params->model_path = "";
+        }
     }
 
     sddebugmode = inputs.debugmode;
@@ -263,31 +308,37 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
         sd_params->control_net_cpu);
     }
 
-    sd_ctx = new_sd_ctx(sd_params->model_path.c_str(),
-                        sd_params->clip_l_path.c_str(),
-                        sd_params->clip_g_path.c_str(),
-                        sd_params->t5xxl_path.c_str(),
-                        sd_params->diffusion_model_path.c_str(),
-                        sd_params->vae_path.c_str(),
-                        sd_params->taesd_path.c_str(),
-                        sd_params->controlnet_path.c_str(),
-                        sd_params->lora_model_dir.c_str(),
-                        sd_params->embeddings_path.c_str(),
-                        sd_params->stacked_id_embeddings_path.c_str(),
-                        vae_decode_only,
-                        sd_params->vae_tiling,
-                        free_param,
-                        sd_params->n_threads,
-                        sd_params->wtype,
-                        sd_params->rng_type,
-                        sd_params->schedule,
-                        sd_params->clip_on_cpu,
-                        sd_params->control_net_cpu,
-                        sd_params->vae_on_cpu,
-                        sd_params->diffusion_flash_attn,
-                        sd_params->chroma_use_dit_mask,
-                        sd_params->chroma_use_t5_mask,
-                        sd_params->chroma_t5_mask_pad);
+    sd_ctx_params_t params;
+    sd_ctx_params_init(&params);
+    params.model_path = sd_params->model_path.c_str();
+    params.clip_l_path = sd_params->clip_l_path.c_str();
+    params.clip_g_path = sd_params->clip_g_path.c_str();
+    params.t5xxl_path = sd_params->t5xxl_path.c_str();
+    params.diffusion_model_path = sd_params->diffusion_model_path.c_str();
+    params.vae_path = sd_params->vae_path.c_str();
+    params.taesd_path = sd_params->taesd_path.c_str();
+    params.control_net_path = sd_params->controlnet_path.c_str();
+    params.lora_model_dir = sd_params->lora_model_dir.c_str();
+    params.embedding_dir = sd_params->embeddings_path.c_str();
+    params.stacked_id_embed_dir = sd_params->stacked_id_embeddings_path.c_str();
+    params.vae_decode_only = vae_decode_only;
+    params.vae_tiling = sd_params->vae_tiling;
+    params.free_params_immediately = free_param;
+    params.n_threads = sd_params->n_threads;
+    params.wtype = sd_params->wtype;
+    params.rng_type = sd_params->rng_type;
+    params.schedule = sd_params->schedule;
+    params.keep_clip_on_cpu = sd_params->clip_on_cpu;
+    params.keep_control_net_on_cpu = sd_params->control_net_cpu;
+    params.keep_vae_on_cpu = sd_params->vae_on_cpu;
+    params.diffusion_flash_attn = sd_params->diffusion_flash_attn;
+    params.diffusion_conv_direct = sd_params->diffusion_conv_direct;
+    params.vae_conv_direct = sd_params->vae_conv_direct;
+    params.chroma_use_dit_mask = sd_params->chroma_use_dit_mask;
+    params.chroma_use_t5_mask = sd_params->chroma_use_t5_mask;
+    params.chroma_t5_mask_pad = sd_params->chroma_t5_mask_pad;
+
+    sd_ctx = new_sd_ctx(&params);
 
     if (sd_ctx == NULL) {
         printf("\nError: KCPP SD Failed to create context!\nIf using Flux/SD3.5, make sure you have ALL files required (e.g. VAE, T5, Clip...) or baked in!\n");
@@ -301,7 +352,6 @@ bool sdtype_load_model(const sd_load_model_inputs inputs) {
     if(lorafilename!="" && inputs.lora_multiplier>0)
     {
         printf("\nApply LoRA...\n");
-       // sd_ctx->sd->set_pending_lora(lorafilename,inputs.lora_multiplier);
         sd_ctx->sd->apply_lora_from_file(lorafilename,inputs.lora_multiplier);
     }
 
@@ -478,11 +528,29 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
     auto loadedsdver = get_loaded_sd_version(sd_ctx);
     if (loadedsdver == SDVersion::VERSION_FLUX)
     {
-        if (!sd_loaded_chroma()) {
-            sd_params->cfg_scale = 1;  //non chroma clamp cfg scale
+        if (loaded_model_is_chroma(sd_ctx)) {
+            if (sd_params->diffusion_flash_attn && sd_params->chroma_use_dit_mask) {
+                if (!sd_is_quiet && sddebugmode) {
+                    printf("Chroma: flash attention is on, disabling DiT mask\n");
+                }
+                sd_params->chroma_use_dit_mask = false;
+            }
+        }
+        else {
+            if (sd_params->cfg_scale != 1.0f) {
+                //non chroma clamp cfg scale
+                if (!sd_is_quiet && sddebugmode) {
+                    printf("Flux: clamping CFG Scale to 1\n");
+                }
+                sd_params->cfg_scale = 1.0f;
+            }
         }
         if (sampler == "euler a" || sampler == "k_euler_a" || sampler == "euler_a") {
-            sampler = "euler";  //euler a broken on flux
+            //euler a broken on flux
+            if (!sd_is_quiet && sddebugmode) {
+                printf("Flux: switching Euler A to Euler\n");
+            }
+            sampler = "euler";
         }
     }
 
@@ -517,17 +585,6 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
     bool dotile = (sd_params->width*sd_params->height > cfg_tiled_vae_threshold*cfg_tiled_vae_threshold);
     set_sd_vae_tiling(sd_ctx,dotile); //changes vae tiling, prevents memory related crash/oom
 
-    if (sd_params->clip_skip <= 0) {
-        // workaround for clip_skip being "stuck" at the previous requested value
-        // 2 is the default for all recent base models (SD2, SDXL, Flux, SD3)
-        if (sd_version_is_sd1((SDVersion)loadedsdver)) {
-            sd_params->clip_skip = 1;
-        }
-        else {
-            sd_params->clip_skip = 2;
-        }
-    }
-
     //for img2img
     sd_image_t input_image = {0,0,0,nullptr};
     std::vector<sd_image_t> extraimage_references;
@@ -659,25 +716,25 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
         }
     }
 
-    std::vector<sd_image_t> kontext_imgs;
-    if(extra_image_data.size()>0 && loadedsdver==SDVersion::VERSION_FLUX && !sd_loaded_chroma())
+    std::vector<sd_image_t> reference_imgs;
+    if(extra_image_data.size()>0 && loadedsdver==SDVersion::VERSION_FLUX && !loaded_model_is_chroma(sd_ctx))
     {
         for(int i=0;i<extra_image_data.size();++i)
         {
-            kontext_imgs.push_back(extraimage_references[i]);
+            reference_imgs.push_back(extraimage_references[i]);
         }
         if(!sd_is_quiet && sddebugmode==1)
         {
-            printf("\nFlux Kontext: Using %d reference images\n",kontext_imgs.size());
+            printf("\nFlux Kontext: Using %d reference images\n",reference_imgs.size());
         }
     }
 
-    std::vector<sd_image_t*> photomaker_imgs;
+    std::vector<sd_image_t> photomaker_imgs;
     if(photomaker_enabled && extra_image_data.size()>0)
     {
         for(int i=0;i<extra_image_data.size();++i)
         {
-            photomaker_imgs.push_back(&extraimage_references[i]);
+            photomaker_imgs.push_back(extraimage_references[i]);
         }
         if(!sd_is_quiet && sddebugmode==1)
         {
@@ -685,6 +742,41 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
         }
     }
 
+    sd_img_gen_params_t params;
+    sd_img_gen_params_init (&params);
+
+    params.prompt = sd_params->prompt.c_str();
+    params.negative_prompt = sd_params->negative_prompt.c_str();
+    params.clip_skip = sd_params->clip_skip;
+    params.guidance.txt_cfg = sd_params->cfg_scale;
+    params.guidance.img_cfg = sd_params->cfg_scale;
+    params.guidance.distilled_guidance = sd_params->guidance;
+    params.eta = sd_params->eta;
+    params.width = sd_params->width;
+    params.height = sd_params->height;
+    params.sample_method = sd_params->sample_method;
+    params.sample_steps = sd_params->sample_steps;
+    params.seed = sd_params->seed;
+    params.batch_count = sd_params->batch_count;
+    params.control_cond = control_image;
+    params.control_strength = sd_params->control_strength;
+    params.style_strength = sd_params->style_ratio;
+    params.normalize_input = sd_params->normalize_input;
+    params.input_id_images_path = sd_params->input_id_images_path.c_str();
+
+    params.guidance.slg.layers = sd_params->skip_layers.data();
+    params.guidance.slg.layer_count = sd_params->skip_layers.size();
+    params.guidance.slg.layer_start = sd_params->skip_layer_start;
+    params.guidance.slg.layer_end = sd_params->skip_layer_end;
+    params.guidance.slg.scale = sd_params->slg_scale;
+
+    params.ref_images = reference_imgs.data();
+    params.ref_images_count = reference_imgs.size();
+
+    kcpp_img_gen_params_t extra_params = {};
+    extra_params.photomaker_references = photomaker_imgs.data();
+    extra_params.photomaker_reference_count = photomaker_imgs.size();
+
     if (sd_params->mode == TXT2IMG) {
 
         if(!sd_is_quiet && sddebugmode==1)
@@ -704,32 +796,8 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
             sd_params->control_strength);
         }
 
+        results = generate_image(sd_ctx, &params, &extra_params);
 
-        results = txt2img(sd_ctx,
-                          sd_params->prompt.c_str(),
-                          sd_params->negative_prompt.c_str(),
-                          sd_params->clip_skip,
-                          sd_params->cfg_scale,
-                          sd_params->guidance,
-                          sd_params->eta,
-                          sd_params->width,
-                          sd_params->height,
-                          sd_params->sample_method,
-                          sd_params->sample_steps,
-                          sd_params->seed,
-                          sd_params->batch_count,
-                          control_image,
-                          sd_params->control_strength,
-                          sd_params->style_ratio,
-                          sd_params->normalize_input,
-                          sd_params->input_id_images_path.c_str(),
-                          kontext_imgs.data(), kontext_imgs.size(),
-                          sd_params->skip_layers.data(),
-                          sd_params->skip_layers.size(),
-                          sd_params->slg_scale,
-                          sd_params->skip_layer_start,
-                          sd_params->skip_layer_end,
-                          photomaker_imgs);
     } else {
 
         if (sd_params->width <= 0 || sd_params->width % 64 != 0 || sd_params->height <= 0 || sd_params->height % 64 != 0) {
@@ -835,34 +903,12 @@ sd_generation_outputs sdtype_generate(const sd_generation_inputs inputs)
             sd_params->strength);
         }
 
-        results = img2img(sd_ctx,
-                            input_image,
-                            mask_image,
-                            sd_params->prompt.c_str(),
-                            sd_params->negative_prompt.c_str(),
-                            sd_params->clip_skip,
-                            sd_params->cfg_scale,
-                            sd_params->guidance,
-                            sd_params->eta,
-                            sd_params->width,
-                            sd_params->height,
-                            sd_params->sample_method,
-                            sd_params->sample_steps,
-                            sd_params->strength,
-                            sd_params->seed,
-                            sd_params->batch_count,
-                            control_image,
-                            sd_params->control_strength,
-                            sd_params->style_ratio,
-                            sd_params->normalize_input,
-                            sd_params->input_id_images_path.c_str(),
-                            kontext_imgs.data(), kontext_imgs.size(),
-                            sd_params->skip_layers.data(),
-                            sd_params->skip_layers.size(),
-                            sd_params->slg_scale,
-                            sd_params->skip_layer_start,
-                            sd_params->skip_layer_end,
-                            photomaker_imgs);
+        params.strength = sd_params->strength;
+        params.init_image = input_image;
+        params.mask_image = mask_image;
+
+        results = generate_image(sd_ctx, &params, &extra_params);
+
     }
 
     if (results == NULL) {
